import logging
from itertools import starmap, chain

from bson import DEFAULT_CODEC_OPTIONS
from frontera.contrib.scrapy.middlewares.seeds import SeedLoader
from frontera.exceptions import NotConfigured
from pymongo import MongoClient
from pymongo.cursor import Cursor
from scrapy.crawler import Crawler
from scrapy.http import Request
from scrapy.settings import Settings
from scrapy.signals import spider_closed
from scrapy.signals import spider_opened
from scrapy.spiders import Spider
from scrapy.utils.misc import load_object

from frontera_seedloader_mongodb.settings.default_settings import \
    SEEDS_MONGODB_COLLECTION
from frontera_seedloader_mongodb.settings.default_settings import \
    SEEDS_MONGODB_DATABASE
from frontera_seedloader_mongodb.settings.default_settings import \
    SEEDS_MONGODB_SEEDS_BATCH_SIZE
from frontera_seedloader_mongodb.settings.default_settings import \
    SEEDS_MONGODB_SEEDS_PREPARE
from frontera_seedloader_mongodb.settings.default_settings import \
    SEEDS_MONGODB_SEEDS_QUERY
from frontera_seedloader_mongodb.utils.get_mongodb_uri import get_mongodb_uri

logger = logging.getLogger(__name__)


class MongoDBSeedLoader(SeedLoader):
    def configure(self, settings: Settings):
        self.settings = settings
        self.uri = get_mongodb_uri(self.settings)
        self.codec_options = DEFAULT_CODEC_OPTIONS.with_options(
            unicode_decode_error_handler='ignore')

    @classmethod
    def from_crawler(cls, crawler: Crawler):
        obj = cls(crawler)
        crawler.signals.connect(obj.open_spider, signal=spider_opened)
        crawler.signals.connect(obj.close_spider, signal=spider_closed)
        return obj

    def open_spider(self, spider: Spider):
        try:
            if self.settings.get(SEEDS_MONGODB_SEEDS_PREPARE):
                self.prepare = load_object(
                    self.settings.get(SEEDS_MONGODB_SEEDS_PREPARE))
            else:
                self.prepare = lambda x: map(
                    lambda y: (y, {'seed': x}),
                    x['websites'])
        except:
            raise NotConfigured

        self.cnx = MongoClient(self.uri)
        self.db = self.cnx.get_database(
            self.settings.get(SEEDS_MONGODB_DATABASE, 'seeds'))
        self.coll = self.db.get_collection(
            self.settings.get(SEEDS_MONGODB_COLLECTION, 'seeds'))

        logger.info('Spider opened: Open the connection to MongoDB: %s',
                    self.uri)

    def close_spider(self, spider: Spider):
        self.cnx.close()
        logger.info('Spider closed: Close the connection to MongoDB %s',
                    self.uri)

    def process_start_requests(self, start_requests, spider: Spider):
        yield from starmap(
            lambda x, y: Request(url=x, meta=y),
            chain(*map(self.prepare, self.load_seeds())))

    def load_seeds(self) -> Cursor:
        logger.info(
            'There are %d seeds.',
            self.coll.count(
                self.settings.get(
                    SEEDS_MONGODB_SEEDS_QUERY, {}
                ).get('filter')
            ))
        return self.coll.find(
            **self.settings.get(SEEDS_MONGODB_SEEDS_QUERY, {})
        ).batch_size(
            self.settings.get(SEEDS_MONGODB_SEEDS_BATCH_SIZE, 100)
        )



import logging
import pprint
from typing import Generator

from bson import DEFAULT_CODEC_OPTIONS
from frontera import Settings
from frontera.contrib.scrapy.middlewares.seeds import SeedLoader
from scrapy import Request
from scrapy.crawler import Crawler
from scrapy.signals import spider_closed
from scrapy.signals import spider_opened
from scrapy.spiders import Spider
from twisted.internet.defer import inlineCallbacks, Deferred
from txmongo.connection import ConnectionPool

from .....settings.default_settings import SEEDS_MONGODB_COLLECTION
from .....settings.default_settings import SEEDS_MONGODB_DATABASE
from .....utils.get_mongodb_uri import get_mongodb_uri

logger = logging.getLogger(__name__)
pp = pprint.PrettyPrinter(indent=4)


class MongoDBAsyncSeedLoader(SeedLoader):
    def configure(self, settings: Settings):
        self.settings = self.crawler.settings
        self.uri = get_mongodb_uri(self.settings)
        self.codec_options = DEFAULT_CODEC_OPTIONS.with_options(
            unicode_decode_error_handler='ignore')

    @classmethod
    def from_crawler(cls, crawler: Crawler):
        obj = cls(crawler)
        crawler.signals.connect(obj.open_spider, signal=spider_opened)
        crawler.signals.connect(obj.close_spider, signal=spider_closed)
        return obj

    @inlineCallbacks
    def open_spider(self, spider: Spider):
        self.cnx = yield ConnectionPool(
            self.uri,
            codec_options=self.codec_options)
        self.db = yield getattr(
            self.cnx,
            self.settings.get(SEEDS_MONGODB_DATABASE, 'seeds'))
        self.coll = yield getattr(
            self.db,
            self.settings.get(SEEDS_MONGODB_COLLECTION, 'seeds'))

        yield self.coll.with_options(codec_options=self.codec_options)

        logger.info('Spider opened: Open the connection to MongoDB: %s',
                    self.uri)

    @inlineCallbacks
    def close_spider(self, spider: Spider):
        yield self.cnx.disconnect()
        logger.info('Spider closed: Close the connection to MongoDB %s',
                    self.uri)

    def process_start_requests(
            self,
            start_requests,
            spider: Spider) -> Generator[Request, None, None]:
        spider.start_urls = self.load_seeds
        yield from spider.start_requests()

    @inlineCallbacks
    def load_seeds(self, dfr: Deferred = None):
        if not dfr:
            docs, dfr = yield self.coll.find_with_cursor(
                # **self.settings.get(SEEDS_MONGODB_SEEDS_QUERY, {'filter': {}})
                cursor=True
            )
            return docs, dfr
        else:
            docs, dfr = yield dfr
            return docs, dfr
            # while docs:
            #     for doc in docs:
            #         yield doc
            #     else:
            #         docs, dfr = yield dfr



SEEDS_MONGODB_USERNAME = 'SEEDS_MONGODB_USERNAME'
SEEDS_MONGODB_PASSWORD = 'SEEDS_MONGODB_PASSWORD'

SEEDS_MONGODB_HOST = 'SEEDS_MONGODB_HOST'
SEEDS_MONGODB_PORT = 'SEEDS_MONGODB_PORT'

SEEDS_MONGODB_DATABASE = 'SEEDS_MONGODB_DATABASE'
SEEDS_MONGODB_COLLECTION = 'SEEDS_MONGODB_COLLECTION'

SEEDS_MONGODB_OPTIONS_ = 'SEEDS_MONGODB_OPTIONS_'

SEEDS_MONGODB_SEEDS_QUERY = 'SEEDS_MONGODB_SEEDS_QUERY'
SEEDS_MONGODB_SEEDS_BATCH_SIZE = 'SEEDS_MONGODB_SEEDS_BATCH_SIZE'

SEEDS_MONGODB_SEEDS_PREPARE = 'SEEDS_MONGODB_SEEDS_PREPARE'



from scrapy.settings import Settings

from ..settings.default_settings import SEEDS_MONGODB_DATABASE
from ..settings.default_settings import SEEDS_MONGODB_HOST
from ..settings.default_settings import SEEDS_MONGODB_OPTIONS_
from ..settings.default_settings import SEEDS_MONGODB_PASSWORD
from ..settings.default_settings import SEEDS_MONGODB_PORT
from ..settings.default_settings import SEEDS_MONGODB_USERNAME


def get_mongodb_uri(settings: Settings) -> str:
    return 'mongodb://{account}{path}{options}'.format(
        account=_gen_mongo_account(settings),
        path=_gen_mongo_path(settings),
        options=_gen_mongo_option(settings))


def _gen_mongo_account(settings: Settings) -> str:
    return {
        True: '{username}:{password}@'.format(
            username=settings.get(SEEDS_MONGODB_USERNAME),
            password=settings.get(SEEDS_MONGODB_PASSWORD)),
        False: ''
    }.get(all(map(settings.get, [SEEDS_MONGODB_USERNAME, SEEDS_MONGODB_PASSWORD])))


def _gen_mongo_path(settings: Settings) -> str:
    return '{host}:{port}/{database}'.format(
        host=settings.get(SEEDS_MONGODB_HOST, 'localhost'),
        port=settings.get(SEEDS_MONGODB_PORT, 27017),
        database=settings.get(SEEDS_MONGODB_DATABASE, 'seeds'))


def _gen_mongo_option(settings: Settings) -> str:
    options = list(filter(
        lambda x: all([x[0].startswith(SEEDS_MONGODB_OPTIONS_),
                       x[0].replace(SEEDS_MONGODB_OPTIONS_, '')]),
        settings.attributes.items()
    ))
    if options:
        return '?{options}'.format(
            options='&'.join(map(
                lambda x: '{option}={value}'.format(
                    option=x[0].replace(SEEDS_MONGODB_OPTIONS_, '').lower(),
                    value=x[1]),
                options)))
    else:
        return ''



from setuptools import find_packages
from setuptools import setup

import versioneer

extras_require = {}

setup(
    name='frontera-seedloader-mongodb',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    url='https://github.com/grammy-jiang/frontera-seedloader-mongodb',
    description='',
    long_description=open('README.rst').read(),
    author='Grammy Jiang',
    author_mail='grammy.jiang@gmail.com',
    maintainer='Grammy Jiang',
    maintainer_email='grammy.jiang@gmail.com',
    license='BSD',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Framework :: Scrapy',
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Plugins',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'frontera',
        'scrapy>=1.4.0',
        'txmongo',
    ],
    extras_require=extras_require,
)
